package de.unikoblenz.fgbks.base.utils.boundary;

import de.unikoblenz.fgbks.base.utils.boundary.bicreater.BoundaryBiCreater;
import de.unikoblenz.fgbks.base.utils.boundary.bicreater.BoundaryBiCreaterType;
import de.unikoblenz.fgbks.base.utils.boundary.checker.BoundaryCheck;
import de.unikoblenz.fgbks.base.utils.boundary.checker.BoundaryCheckType;
import de.unikoblenz.fgbks.base.utils.boundary.creater.BoundaryCreater;
import de.unikoblenz.fgbks.base.utils.boundary.creater.BoundaryCreaterType;
import de.unikoblenz.fgbks.base.utils.boundary.impl.InvalidBoundary;
import de.unikoblenz.fgbks.core.dmn.domain.vdmn.VTypeRef;
import java.util.HashMap;
import java.util.Map;
import java.util.Optional;
import org.apache.commons.lang3.Validate;

public abstract class AbstractBoundary<T extends Comparable<? super T>> implements Boundary {

  private HashMap<BoundaryCheckType, BoundaryCheck> checkFunctions;
  private HashMap<BoundaryBiCreaterType, BoundaryBiCreater> biCreaterFunctions;
  private HashMap<BoundaryCreaterType, BoundaryCreater> createrFunctions;
  private Class<T> type;
  protected String text;

  protected AbstractBoundary(String text, Class<T> type) {
    this.text = text;
    this.type = type;
    parse(text);
    validate();
    checkFunctions = new HashMap<>(getChecker());
    biCreaterFunctions = new HashMap<>(getBiCreater());
    createrFunctions = new HashMap<>(getCreater());
  }

  protected AbstractBoundary(Class<T> type) {
    this("", type);
  }

  @Override
  public String getText() {
    return text;
  }

  @Override
  public VTypeRef getTypeRef() {
    return VTypeRef.getTypeRefFromJavaClass(type);
  }

  @Override
  public boolean checkWith(BoundaryCheckType checkType, Boundary other) {
    if (other instanceof InvalidBoundary) {
      return false;
    }
    BoundaryCheck checkFunction = getCheckFunction(checkType);
    if (checkFunction == null || !isValidOperation(checkType, other)) {
      throw new BoundaryOperationNotSupportedException(); // TODO: add error text
    }
    return checkFunction.check(this, other); // TODO, remove RAW type
  }

  @Override
  public Optional<Boundary> create(BoundaryCreaterType createrType) {
    BoundaryCreater creater = getCreaterFunction(createrType);
    if (creater == null || !isValidOperation(createrType)) {
      throw new BoundaryOperationNotSupportedException(); // TODO: add error text
    }
    return creater.create(this);
  }

  @Override
  public Optional<Boundary> createBi(BoundaryBiCreaterType biCreaterType, Boundary other) {
    if (other instanceof InvalidBoundary) {
      return Optional.empty();
    }
    BoundaryBiCreater biCreater = getBiCreaterFunction(biCreaterType);
    if (biCreater == null || !isValidOperation(biCreaterType, other)) {
      throw new BoundaryOperationNotSupportedException(); // TODO: add error text
    }
    return biCreater.create(this, other); // TODO, remove RAW type
  }

  public boolean isValidOperation(BoundaryCheckType checkType, Boundary other) {
    return this.getTypeRef() == other.getTypeRef() && isValidOperation(checkType);
  }

  public boolean isValidOperation(BoundaryCheckType checkType) {
    return checkFunctions.containsKey(Validate.notNull(checkType));
  }

  public boolean isValidOperation(BoundaryBiCreaterType combineType, Boundary other) {
    return this.getTypeRef() == other.getTypeRef() && isValidOperation(combineType);
  }

  public boolean isValidOperation(BoundaryBiCreaterType combineTyp) {
    return biCreaterFunctions.containsKey(Validate.notNull(combineTyp));
  }

  public boolean isValidOperation(BoundaryCreaterType createrType) {
    return createrFunctions.containsKey(Validate.notNull(createrType));
  }

  protected BoundaryCheck getCheckFunction(BoundaryCheckType checkType) {
    return checkFunctions.get(checkType);
  }

  protected BoundaryCreater getCreaterFunction(BoundaryCreaterType createrType) {
    return createrFunctions.get(createrType);
  }

  protected BoundaryBiCreater getBiCreaterFunction(BoundaryBiCreaterType biCreaterTypeType) {
    return biCreaterFunctions.get(biCreaterTypeType);
  }

  protected abstract void parse(String text);

  protected abstract void validate();

  protected abstract Map<BoundaryCheckType, BoundaryCheck> getChecker();

  protected abstract Map<BoundaryBiCreaterType, BoundaryBiCreater> getBiCreater();

  protected abstract Map<BoundaryCreaterType, BoundaryCreater> getCreater();

  @Override
  public String toString() {
    return "{" + getText() + "},{" + getParsedText() + "}";
  }
}
